{-# OPTIONS --type-in-type
    #-}

open import NeilPrelude

module Arrow (Arr : Set -> Set -> Set)

       (pure   : {A B : Set} -> (A -> B) -> Arr A B) 
       (_>>>_  : {A B C : Set} -> Arr A B -> Arr B C -> Arr A C)
       (first  : {A B C : Set} -> Arr A B -> Arr (A × C) (B × C))

-- arrow laws

       (1stLaw : {A B : Set} -> {f : Arr A B} -> (pure id >>> f) ≡ f)
--       (2ndLaw : {A B : Set} -> {f : Arr A B} -> (f >>> pure id) ≡ f) -- Proved Redundant in Arrow Calculus Paper
       (3rdLaw : {A B C D : Set} -> {f : Arr A B} -> {g : Arr B C} -> {h : Arr C D} -> ((f >>> g) >>> h) ≡ (f >>> (g >>> h)))
       (4thLaw : {A B C : Set} -> {f : A -> B} -> {g : B -> C} -> pure (g ∘ f) ≡ pure f >>> pure g)
       (5thLaw : {A B C : Set} -> {f : A -> B} -> first (pure f) ≡ pure (f ∥ id {C}))
       (6thLaw : {A B C D : Set} -> {f : Arr A B} -> {g : Arr B C} -> first {_} {_} {D} (f >>> g) ≡ first f >>> first g)
       (7thLaw : {A B C D : Set} -> {f : Arr A B} -> {g : C -> D} -> first f >>> pure (id ∥ g) ≡ pure (id ∥ g) >>> first f)
       (8thLaw : {A B C : Set} -> {f : Arr A B} -> first {_} {_} {C} f >>> pure fst ≡ pure fst >>> f)
       (9thLaw : {A B C D : Set} -> {f : Arr A B} -> first {A × C} {_} {D} (first f) >>> pure ×assoc ≡ pure ×assoc >>> first f)

where

----------------------------------------------

infixr 90 _⋙_
infixr 91 _***_
infix  91 _&&&_

_⋙_  : {A B C : Set} -> Arr A B -> Arr B C -> Arr A C
_⋙_  = _>>>_

_≪_  : {A B C : Set} -> Arr B C -> Arr A B -> Arr A C
_≪_  = flip _⋙_

arr : {A B : Set} -> (A -> B) -> Arr A B
arr = pure

frst : {A B C : Set} -> Arr A B -> Arr (A × C) (B × C)
frst = first

secnd : {A B C : Set} -> Arr B C -> Arr (A × B) (A × C)
secnd a = pure swap ⋙ first a ⋙ pure swap

_***_   : {A B C D : Set} -> Arr A C -> Arr B D -> Arr (A × B) (C × D)
a *** b = frst a ⋙ secnd b

_&&&_ : {A B C : Set} -> Arr A B -> Arr A C -> Arr A (B × C)
a &&& b = pure dup ⋙ a *** b

arr2 : {A B C : Set} -> (A -> B -> C) -> Arr (A × B) C
arr2 f = arr (uncurry f)

liftA : {A B : Set} -> (A -> B) -> Arr A B
liftA = pure

liftA2 : {A B C D : Set} -> (B -> C -> D) -> Arr A B -> Arr A C -> Arr A D
liftA2 f a b = a &&& b >>> arr2 f